﻿/******************************************************************************
 * Program.cs
 * 
 * This module implements the Program class and related types.
 *  
 * Date:   5/2011
 * 
 * Copyright (c) 2010, 2011, Mark Betz 
 * 
 * All rights reserved. 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 * 
 *   * Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer. 
 *   * Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution. 
 *   * Neither the name of the Author nor the names of contributors may be 
 *     used to endorse or promote products derived from this software
 *     without specific prior written permission. 
 *     
 * THIS SOFTWARE IS PROVIDED BY MARK BETZ ''AS IS'' AND ANY  EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL MARK BETZ BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 */
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace SmoothGen
{
    /// <summary>
    /// Static class container for runtime options
    /// </summary>
    internal static class ShowOptions
    {
        /// <summary>
        /// The title of the slideshow
        /// </summary>
        internal static string Title;
        
        /// <summary>
        /// The width of the slideshow
        /// </summary>
        internal static int Width = 400;
        
        /// <summary>
        /// The height of the slideshow
        /// </summary>
        internal static int Height = 200;
        
        /// <summary>
        /// True if the slideshow should advance automatically on a timer
        /// </summary>
        internal static bool Timed;

        /// <summary>
        /// The delay between slides in milliseconds, valid if Timed == true
        /// </summary>
        internal static int Delay = 5000;
        
        /// <summary>
        /// True if the slideshow should display the prev/next arrows
        /// </summary>
        internal static bool Arrows = true;

        /// <summary>
        /// True if the slideshow should display the thumbnail carousel
        /// </summary>
        internal static bool Thumbs = true;

        /// <summary>
        /// Width of the thumbnails in pixels, valid if Thumbs == true
        /// </summary>
        internal static int ThumbWidth = 100;

        /// <summary>
        /// Height of the thumbnails in pixels, valid if Thumbs == true
        /// </summary>
        internal static int ThumbHeight = 75;
        
        /// <summary>
        /// True if the slideshow should display the information pane
        /// </summary>
        internal static bool Info = true;

        /// <summary>
        /// Height of the information pane, valid if Info == true
        /// </summary>
        internal static int InfoHeight = 40;

        /// <summary>
        /// True if the slideshow should wrap images in links
        /// </summary>
        internal static bool Links = false;

        /// <summary>
        /// True if the slideshow should use the history manager
        /// </summary>
        internal static bool History = false;

        /// <summary>
        /// True if slide titles should be imported from image title metadata
        /// </summary>
        internal static bool ImportTitles = true;

        /// <summary>
        /// True if slide descriptions should be imported from image comment metadata
        /// </summary>
        internal static bool ImportComments = true;
        
        /// <summary>
        /// The base url to prepend to slide img tag src attribute values
        /// </summary>
        internal static string ImageBaseUrl;

        /// <summary>
        /// The base url to prepend to script tag src attribute values
        /// </summary>
        internal static string ScriptBaseUrl;

        /// <summary>
        /// The base url to prepend to style link addresses
        /// </summary>
        internal static string StyleBaseUrl;
    }

    class Program
    {
        const string BadArgumentFormatMessage = "Malformed command line argument.";
        const string BadArgumentValueMessage = "Bad value in command line argument: {0}.";
        const string BadArgumentMessage = "Unrecognized command line argument: {0}.";

        /// <summary>
        /// Parses the command line arguments and updates the ShowOptions field values
        /// </summary>
        /// <param name="args"></param>
        private static void UpdateShowOptions(string[] args)
        {
            for (int i = 2; i < args.Length; i++)
            {
                string[] parts = args[i].Split(new char[] {'='});
                if (parts.Length < 2)
                    throw new ApplicationException(BadArgumentFormatMessage);

                int numBucket;
                bool boolBucket;

                switch (parts[0])
                {
                    case "width" :
                        if (Int32.TryParse(parts[1], out numBucket))
                            ShowOptions.Width = numBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "width"));
                        break;

                    case "height" :
                        if (Int32.TryParse(parts[1], out numBucket))
                            ShowOptions.Height = numBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "height"));
                        break;

                    case "timed" :
                        if (Boolean.TryParse(parts[1], out boolBucket))
                            ShowOptions.Timed = boolBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "timed"));
                        break;

                    case "delay":
                        if (Int32.TryParse(parts[1], out numBucket))
                            ShowOptions.Delay = numBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "delay"));
                        break;

                    case "arrows":
                        if (Boolean.TryParse(parts[1], out boolBucket))
                            ShowOptions.Arrows = boolBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "arrows"));
                        break;

                    case "thumbs":
                        if (Boolean.TryParse(parts[1], out boolBucket))
                            ShowOptions.Thumbs = boolBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "thumbs"));
                        break;

                    case "thumbw":
                        if (Int32.TryParse(parts[1], out numBucket))
                            ShowOptions.ThumbWidth = numBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "thumbw"));
                        break;

                    case "thumbh":
                        if (Int32.TryParse(parts[1], out numBucket))
                            ShowOptions.ThumbHeight = numBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "thumbh"));
                        break;

                    case "info":
                        if (Boolean.TryParse(parts[1], out boolBucket))
                            ShowOptions.Info = boolBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "info"));
                        break;

                    case "infoheight":
                        if (Int32.TryParse(parts[1], out numBucket))
                            ShowOptions.InfoHeight = numBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "infoheight"));
                        break;

                    case "links":
                        if (Boolean.TryParse(parts[1], out boolBucket))
                            ShowOptions.Links = boolBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "links"));
                        break;

                    case "history":
                        if (Boolean.TryParse(parts[1], out boolBucket))
                            ShowOptions.History = boolBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "history"));
                        break;

                    case "importtitles":
                        if (Boolean.TryParse(parts[1], out boolBucket))
                            ShowOptions.ImportTitles = boolBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "importtitles"));
                        break;

                    case "importcomments":
                        if (Boolean.TryParse(parts[1], out boolBucket))
                            ShowOptions.ImportComments = boolBucket;
                        else throw new ApplicationException(String.Format(BadArgumentValueMessage, "importcomments"));
                        break;

                    case "imagebaseurl":
                        ShowOptions.ImageBaseUrl = parts[1];
                        if (!ShowOptions.ImageBaseUrl.EndsWith("/"))
                            ShowOptions.ImageBaseUrl += "/";
                        break;

                    case "scriptbaseurl":
                        ShowOptions.ScriptBaseUrl = parts[1];
                        if (!ShowOptions.ScriptBaseUrl.EndsWith("/"))
                            ShowOptions.ScriptBaseUrl += "/";
                        break;

                    case "stylebaseurl":
                        ShowOptions.StyleBaseUrl = parts[1];
                        if (!ShowOptions.StyleBaseUrl.EndsWith("/"))
                            ShowOptions.StyleBaseUrl += "/";
                        break;

                    default:
                        throw new ApplicationException(String.Format(BadArgumentMessage, parts[0]));
                }
            }
        }

        /// <summary>
        /// Applies the current values of the ShowOptions fields to the appropriate tags
        /// in the html template
        /// </summary>
        /// <param name="outDoc"></param>
        private static void ApplyShowOptions(XDocument outDoc)
        {
            var title = outDoc.XPathSelectElement("//title");
            title.Value = ShowOptions.Title;

            title = outDoc.XPathSelectElement("//p[@id='galleryTitle']");
            title.Value = ShowOptions.Title;

            var refstyles = outDoc.XPathSelectElements("//link[@rel='stylesheet']");
            foreach (XElement style in refstyles)
            {
                string href = style.Attribute("href").Value;
                style.Attribute("href").Value = href.Replace("#STYLEBASEURL#", ShowOptions.StyleBaseUrl);
            }
            
            var scripts = outDoc.XPathSelectElements("//script");
            foreach (XElement script in scripts)
            {
                if (null != script.Attribute("src"))
                {
                    string src = script.Attribute("src").Value;
                    script.Attribute("src").Value = src.Replace("#SCRIPTBASEURL#", ShowOptions.ScriptBaseUrl);
                }
            }

            var styles = outDoc.XPathSelectElements("//style[@type='text/css']");
            foreach (XElement style in styles)
            {
                if (style.Value.Contains("#myGallery"))
                {
                    string rule = style.Value;
                    style.Value = rule.Replace("#WIDTH#", ShowOptions.Width.ToString())
                    .Replace("#HEIGHT#", ShowOptions.Height.ToString())
                    .Replace("#CAROUSELCONTAINERHEIGHT#", (ShowOptions.ThumbHeight + 60).ToString())
                    .Replace("#CAROUSELHEIGHT#", (ShowOptions.ThumbHeight + 40).ToString())
                    .Replace("#CAROUSELWRAPPERHEIGHT#", (ShowOptions.ThumbHeight + 3).ToString())
                    .Replace("#INFOHEIGHT#", ShowOptions.InfoHeight.ToString());
                    break;
                }
            }

            XElement startGalleryScript = outDoc.XPathSelectElement("//script[@id='startgallery']");
            if (null != startGalleryScript)
            {
                string code = startGalleryScript.Value;
                startGalleryScript.Value = 
                    code.Replace("#TIMED#", ShowOptions.Timed ? "true" : "false")
                    .Replace("#DELAY#", ShowOptions.Delay.ToString())
                    .Replace("#SHOWARROWS#", ShowOptions.Arrows ? "true" : "false")
                    .Replace("#SHOWCAROUSEL#", ShowOptions.Thumbs ? "true" : "false")
                    .Replace("#SHOWINFOPANE#", ShowOptions.Info ? "true" : "false")
                    .Replace("#EMBEDLINKS#", ShowOptions.Links ? "true" : "false")
                    .Replace("#USEHISTORY#", ShowOptions.History ? "true" : "false")
                    .Replace("#HISTORYMANAGERSTARTCALL#", ShowOptions.History ? "HistoryManager.start();" : String.Empty)
                    .Replace("#THUMBWIDTH#", ShowOptions.ThumbWidth.ToString())
                    .Replace("#THUMBHEIGHT#", ShowOptions.ThumbHeight.ToString());
            }
            else throw new ApplicationException("Gallery startup script block missing.");
        }

        /// <summary>
        /// Returns a FileInfo representing the path and filename for a thumbnail
        /// </summary>
        /// <param name="image"></param>
        /// <returns></returns>
        private static FileInfo ThumbInfo(FileInfo image)
        {
            string imageBaseName = image.FullName.Split(new char[] { '.' })[0];
            return new FileInfo(Path.Combine(image.DirectoryName, imageBaseName + "_t" + image.Extension));
        }

        /// <summary>
        /// Creates and returns an image element, also creates the thumbnail image if needed,
        /// and imports the metadata if needed
        /// </summary>
        /// <param name="image"></param>
        /// <returns></returns>
        private static XElement CreateImageElement(FileInfo image)
        {
            BitmapSource imageSource = BitmapFrame.Create(new Uri(image.FullName));
            string title = String.Empty;
            string description = String.Empty;
            if (ShowOptions.ImportTitles || ShowOptions.ImportComments)
            {
                BitmapMetadata meta = (BitmapMetadata)imageSource.Metadata;

                if (ShowOptions.ImportTitles)
                    title = meta.Title;
                if (ShowOptions.ImportComments)
                    description = meta.Comment;
            }
            XElement imageElement = new XElement("div");
            imageElement.Add(new XAttribute("class", "imageElement"));
            imageElement.Add(new XElement("h3", title));
            imageElement.Add(new XElement("p", description ?? ""));
            imageElement.Add(new XElement("img", new XAttribute("class", "full"), new XAttribute("src", ShowOptions.ImageBaseUrl + image.Name)));

            if (ShowOptions.Thumbs)
            {
                FileInfo thumb = ThumbInfo(image);
                if (thumb.Exists)
                    File.Delete(thumb.FullName);
                
                ScaleTransform st = new ScaleTransform();
                st.ScaleX = (double)ShowOptions.ThumbWidth / (double)imageSource.PixelWidth;
                st.ScaleY = (double)ShowOptions.ThumbHeight / (double)imageSource.PixelHeight;
                TransformedBitmap tb = new TransformedBitmap(imageSource, st);
                BitmapMetadata thumbMeta = new BitmapMetadata("jpg");
                thumbMeta.Title = "sg_thumb";
                JpegBitmapEncoder encoder = new JpegBitmapEncoder();
                encoder.QualityLevel = 100;
                encoder.Frames.Add(BitmapFrame.Create(tb, null, thumbMeta, null));
                
                using (FileStream stream = new FileStream(thumb.FullName, FileMode.Create))
                {
                    encoder.Save(stream);
                    stream.Close();
                }

                imageElement.Add(new XElement("img", new XAttribute("class", "thumbnail"), new XAttribute("src", ShowOptions.ImageBaseUrl + thumb.Name)));
            }
            return imageElement;
        }

        /// <summary>
        /// Dumps the current options values to the console
        /// </summary>
        /// <param name="imageFolder"></param>
        /// <param name="outputFile"></param>
        private static void DumpShowOptions(string imageFolder, string outputFile)
        {
            Console.WriteLine("Current slideshow options:");
            Console.WriteLine("Title: {0}", ShowOptions.Title);
            Console.WriteLine("Image folder: {0}", imageFolder);
            Console.WriteLine("Output html: {0}", outputFile);
            Console.WriteLine("Width: {0}", ShowOptions.Width.ToString());
            Console.WriteLine("Height: {0}", ShowOptions.Height.ToString());
            Console.WriteLine("Timed: {0}", ShowOptions.Timed ? "true" : "false");
            Console.WriteLine("Delay: {0}", ShowOptions.Timed ? ShowOptions.Delay.ToString() : "n/a");
            Console.WriteLine("Arrows: {0}", ShowOptions.Arrows ? "true" : "false");
            Console.WriteLine("Thumbs: {0}", ShowOptions.Thumbs ? "true" : "false");
            Console.WriteLine("Thumb width: {0}", ShowOptions.Thumbs ? ShowOptions.ThumbWidth.ToString() : "n/a");
            Console.WriteLine("Thumb height: {0}", ShowOptions.Thumbs ? ShowOptions.ThumbHeight.ToString() : "n/a");
            Console.WriteLine("Info: {0}", ShowOptions.Info ? "true" : "false");
            Console.WriteLine("Info height: {0}", ShowOptions.Info ? ShowOptions.InfoHeight.ToString() : "n/a");
            Console.WriteLine("Links: {0}", ShowOptions.Links ? "true" : "false");
            Console.WriteLine("History: {0}", ShowOptions.History ? "true" : "false");
            Console.WriteLine("Import titles: {0}", ShowOptions.ImportTitles ? "true" : "false");
            Console.WriteLine("Import comments: {0}", ShowOptions.ImportComments ? "true" : "false");
            Console.WriteLine("Image base url: {0}", ShowOptions.ImageBaseUrl);
            Console.WriteLine("Script base url: {0}", ShowOptions.ScriptBaseUrl);
            Console.WriteLine("Style base url: {0}", ShowOptions.StyleBaseUrl);
        }
        
        /// <summary>
        /// Displays the correct program usage if the command line is missing or malformed
        /// </summary>
        private static void DumpUsage()
        {
            Console.WriteLine("Usage: smoothgen imagefolder outputpath [options]\r\n");
            Console.WriteLine("  imagefolder - the path to the folder containing images for the slideshow.");
            Console.WriteLine("  outputpath - the path and filename for the generated .html file.");
            Console.WriteLine("  options - any of the following:");
            Console.WriteLine("    width=400 - sets the width of the slideshow in pixels.");
            Console.WriteLine("    height=200 - sets the height of the slideshow in pixels.");
            Console.WriteLine("    timed=false - if true the slideshow will advance automatically.");
            Console.WriteLine("    delay=5000 - sets the delay between slides in ms when timed = true.");
            Console.WriteLine("    arrows=true - if true the slideshow will display previous/next arrows.");
            Console.WriteLine("    thumbs=true - if true the slideshow will display the thumbnail carousel.");
            Console.WriteLine("    thumbw=100 - sets the width of thumbnail images if thumbs=true.");
            Console.WriteLine("    thumbh=75 - sets the height of thumbnail images if thumbs=true.");
            Console.WriteLine("    info=true - if true the slideshow will display the image info pane.");
            Console.WriteLine("    infoheight=40 - sets the height of the info pane when info = true");
            Console.WriteLine("    links=false - if true the slideshow will embed image links.");
            Console.WriteLine("    history=false - if true the slideshow will use the history manager.");
            Console.WriteLine("    importtitles=true - if true titles are imported from metadata.");
            Console.WriteLine("    importcomments=true - if true descriptions are imported from metadata.");
            Console.WriteLine("    imagebaseurl=url - a url to prepend to all image filenames.");
            Console.WriteLine("    scriptbaseurl=url - a url to prepend to all script filenames.");
            Console.WriteLine("    stylebaseurl=url - a url to prepend to all style filenames.");
        }

        /// <summary>
        /// Filters the image list to make sure that we don't process previously
        /// generated thumbnails
        /// </summary>
        /// <param name="images"></param>
        /// <returns></returns>
        private static FileInfo[] FilterImageList(FileInfo[] images)
        {
            var filtered = new List<FileInfo>();

            foreach (FileInfo image in images)
            { 
                BitmapSource imageSource = BitmapFrame.Create(new Uri(image.FullName));
                BitmapMetadata meta = (BitmapMetadata)imageSource.Metadata;
                if(!meta.Title.Equals("sg_thumb"))
                    filtered.Add(image);
            }
            return filtered.ToArray();
        }

        /// <summary>
        /// Runs the slideshow generator
        /// </summary>
        /// <param name="args"></param>
        private static void RunGenerator(string[] args)
        {
            ConsoleKeyInfo inKey;

            if (args.Length < 2)
            {
                DumpUsage();
                return;
            }

            DirectoryInfo diInput = new DirectoryInfo(args[0]);
            if (!diInput.Exists)
                throw new ApplicationException("Image folder not found");

            DirectoryInfo diOutput = new DirectoryInfo(Path.GetDirectoryName(args[1]));
            if (!diOutput.Exists)
            {
                Console.Write("Output folder not found - Create? (Y/N): ");
                inKey = Console.ReadKey(true);
                Console.WriteLine();
                if (inKey.KeyChar.Equals('Y') || inKey.KeyChar.Equals('y'))
                    Directory.CreateDirectory(args[1]);
                else
                    throw new ApplicationException("No output folder to write to");
            }
            else
            {
                if (File.Exists(args[1]))
                {
                    Console.Write("Output file {0} already exists - overwrite? (Y/N): ", args[1]);
                    inKey = Console.ReadKey(true);
                    Console.WriteLine();
                    if (inKey.KeyChar.Equals('Y') || inKey.KeyChar.Equals('y'))
                        File.Delete(args[1]);
                    else
                        throw new ApplicationException("No output file to write to");
                }
            }

            if (args.Length > 2)
                UpdateShowOptions(args);

            Console.Write("Enter title (press return to add later): ");
            ShowOptions.Title = Console.ReadLine().TrimEnd(Environment.NewLine.ToCharArray());
            
            DumpShowOptions(args[0], args[1]);
            Console.Write("\r\nGenerate slideshow with current options? (Y/N): ");
            inKey = Console.ReadKey(true);
            Console.WriteLine();
            if (!inKey.KeyChar.Equals('Y') && !inKey.KeyChar.Equals('y'))
                return;

            FileInfo[] images = diInput.GetFiles();
            images = FilterImageList(images);
            if (images.Length == 0)
                throw new ApplicationException(String.Format("No images found in folder {0}", diInput.FullName));
            else 
                Console.WriteLine("Generating SmoothGallery with {0} images", images.Length);

            File.Copy("template.html", args[1]);
            XDocument outDoc = XDocument.Load(args[1]);

            ApplyShowOptions(outDoc);

            XElement contentElement = outDoc.XPathSelectElement("//div[@id='myGallery']");
            foreach (FileInfo image in images)
            { 
                XElement imageElement = CreateImageElement(image);
                if (null != imageElement)
                    contentElement.Add(imageElement);
            }
            Console.WriteLine("Saving {0} ...", args[1]);

            XmlWriterSettings xws = new XmlWriterSettings();
            xws.Indent = true;
            xws.OmitXmlDeclaration = true;
            using (FileStream docStream = File.Open(args[1], FileMode.Create))
            using (XmlWriter xw = XmlWriter.Create(docStream, xws)) 
                outDoc.Save(xw);

            Console.WriteLine("Completed");
        }

        static void Main(string[] args)
        {
            try
            {
                Console.WriteLine("SmoothGen 0.9");
                RunGenerator(args);
            }
            catch (Exception ex)
            {
                Console.WriteLine("Error: {0}", ex.Message);
            }

            if (Debugger.IsAttached)
            {
                Console.WriteLine("Press any key to exit...");
                Console.ReadKey();
            }    
        }
    }
}
